/*
   sfthreshold.c

   This file contains functions that glue the generic thresholding2 code to snort.

   dependent files:  sfthd sfxghash sfghash sflsq 
                     util mstring

   Copyright (C) 2003 Sourcefire,Inc.
   Marc Norton

   2003-05-29:
     cmg: Added s_checked variable  --
       when this is 1, the sfthreshold_test will always return the same answer until
       sfthreshold_reset is called

*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mstring.h"
#include "util.h"

#include "sfthd.h"
#include "sfthreshold.h"

#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif

/*
     Data
*/
static int          s_memcap  = 1024 * 1024;
static THD_STRUCT * s_thd     = 0;
static int          s_enabled = 1;
static int          s_checked = 0; /**< have we evaluated this yet? */
static int          s_answer  = 0; /**< what was the last return value? */


/*

     RULE OPTION - Parse Threshold Option parameters for each RULE

     '; threshold: type limit|threshold|both, track by_src|by_dst, count #, seconds [priority #];'

*/
void ParseThreshold2( THDX_STRUCT * thdx, char * s )
{
      int    i = 0;
      char * argv[100], * t;
      int    argc;
      int    count_flag=0;
      int    seconds_flag=0;
      int    type_flag=0;
      int    tracking_flag=0;
      
      if( !s_enabled )
          return ;

      memset( thdx, 0, sizeof(THDX_STRUCT) );

      thdx->priority = -1; /* Make this lower than standalone threshold command defaults ??? */
      
      /* printf("THRESHOLD-RULE: '%s'\n",s);
       */

      /* Parse all of the args - they come in pairs */      
      for( argc=0, t = strtok(s," ,\n"); argc < 100 &&  t;  argc++, t = strtok(0," ,\n") )
      {
          argv[ argc ] = t;           
          /*  printf("Threshold-RuleOption[%d]:%s\n",argc,t); */
      }

      /* Parameter Check - enough args ?*/
      if( argc != 8 )
      {
          /* Fatal incorrect argument count */ 
          FatalError("Threshold-RuleOptionParse: incorrect argument count\n");
      }

      for(i=0;i<argc;i++)
      {
        if( strcmp(argv[i],"count") == 0  )
         {
            i++;
            thdx->count = atoi(argv[i]);
            count_flag++;
         }
        else if( strcmp(argv[i],"seconds") == 0  )
         {
            i++;
            thdx->seconds = atoi(argv[i]);
            seconds_flag++;
         }
        else if( strcmp(argv[i],"type") == 0  )
         {
            i++;
            if( strcmp(argv[i],"limit") == 0 )
            {
               thdx->type = THD_TYPE_LIMIT;
            }
            else if( strcmp(argv[i],"threshold") == 0 )
            {
               thdx->type = THD_TYPE_THRESHOLD;
            }
            else if( strcmp(argv[i],"both") == 0 )
            {
               thdx->type = THD_TYPE_BOTH;
            }
            else
            {
                /* Fatal incorrect threshold type */
                 FatalError("Threshold-RuleOptionParse: incorrect 'type' argument \n");
            }
            type_flag++;
         }
        else if( strcmp(argv[i],"track") == 0  )
         {
            i++;
            if( strcmp(argv[i],"by_src") == 0 )
            {
                thdx->tracking = THD_TRK_SRC;
            }
            else if( strcmp(argv[i],"by_dst") == 0 )
            {
                thdx->tracking = THD_TRK_DST;
            }
            else
            {
                /* Fatal incorrect threshold type */
                 FatalError("Threshold-RuleOptionParse: incorrect tracking type\n");
            }
            tracking_flag++;
         }
        else
         {
            /* Fatal Out Here - Unknow Option */
            FatalError("Threshold-RuleOptionParse: unknown argument \n");
         }
     }

     if( (count_flag + tracking_flag + type_flag + seconds_flag ) != 4 )
     {
         /* Fatal - incorrect argument count */
         FatalError("Threshold-RuleOptionParse: incorrect argument count\n");
     }

}

/*

     Process the 'config threshold: memcap #bytes, option2-name option2-value, ...'

*/
void ProcessThresholdOptions(char *options)
{
      int     i = 0;
      char ** args;
      int     nargs;
      char ** oargs;
      int     noargs;
      
      if( !s_enabled )
          return ;

      args = mSplit(options,",",10,&nargs,0);  /* get rule option pairs */

      /*
        printf("THRESHOLD-CONF: %d option args\n",nargs);
      */
      
      for(i=0;i<nargs;i++)
      {
          /* printf("THRESHOLD-CONF: arg[%d]=%s\n",i,args[i]); */

          oargs = mSplit(options," ",2,&noargs,0);  /* get rule option pairs */

          if( strcmp(oargs[0],"memcap") == 0  )
          {
             s_memcap = atoi(oargs[1]);
             /* printf("THRESHOLD-CONF: memcap=%d\n",s_memcap); */
          }
          else
          {
             /* Fatal Out Here */
             FatalError("Threshold-RuleOptionParse: unknown argument\n");
          }
     }
}

/*
   don't use priority for 3.0
   
   threshold gen_id #, sig_id #, type limit|threshold|both, track by_src|by_dst, (priority #), count #, seconds #

*/
void ParseSFThreshold( FILE * fp, char * rule )
{
     char        **args, **oargs;
     int         nargs, noargs;
     THDX_STRUCT thdx;
     int         count_flag=0;
     int         seconds_flag=0;
     int         type_flag=0;
     int         tracking_flag=0;
//     int         priority_flag=0;
     int         genid_flag=0;
     int         sigid_flag=0;
     int         i;

     memset( &thdx, 0, sizeof(THDX_STRUCT) );

     while( (*rule <= ' ') && (*rule > 0) ) rule++; /* skip whitespace */
     while( (*rule  > ' ') ) rule++;  /* skip 'threshold' */

     args = mSplit(rule,",",15,&nargs,0);  /* get rule option pairs */

     for( i=0; i<nargs; i++ )
     {
         oargs = mSplit(args[i]," ",2,&noargs,0);  /* get rule option pairs */
         
	 if( noargs != 2 )
         {
             FatalError("Threshold Parse: argument pairing error\n");
         }

         if( strcmp(oargs[0],"type")==0 )
         {
            if( strcmp(oargs[1],"limit") == 0 )
            {
               thdx.type = THD_TYPE_LIMIT;
            }
            else if( strcmp(oargs[1],"threshold") == 0 )
            {
               thdx.type = THD_TYPE_THRESHOLD;
            }
            else if( strcmp(oargs[1],"both") == 0 )
            {
               thdx.type = THD_TYPE_BOTH;
            }
            else
            {
                /* Fatal incorrect threshold type */
                 FatalError("Threshold-Parse: incorrect 'type' argument \n");
            }
            type_flag++;
         }

         else if( strcmp(oargs[0],"track")==0 )
         {
            if( strcmp(oargs[1],"by_src") == 0 )
            {
                thdx.tracking = THD_TRK_SRC;
            }
            else if( strcmp(oargs[1],"by_dst") == 0 )
            {
                thdx.tracking = THD_TRK_DST;
            }
            else
            {
                /* Fatal incorrect threshold type */
                 FatalError("Threshold-Parse: incorrect tracking type\n");
            }
            tracking_flag++;
         }
/*
         else if( strcmp(oargs[0],"priority")==0 )
         {
            thdx.priority = atoi(oargs[1]);
            priority_flag++;
         }
*/
	 else if( strcmp(oargs[0],"count")==0 )
         {
            thdx.count = atoi(oargs[1]);
            count_flag++;
         }

         else if( strcmp(oargs[0],"seconds")==0 )
         {
            thdx.seconds = atoi(oargs[1]);
            seconds_flag++;
         }

         else if( strcmp(oargs[0],"gen_id")==0 )
         {
            thdx.gen_id = atoi(oargs[1]);
            sigid_flag++;
         }

         else if( strcmp(oargs[0],"sig_id")==0 )
         {
            thdx.sig_id = atoi(oargs[1]);
            genid_flag++;
         }
         else
         {
             /* Fatal incorrect threshold type */
             FatalError("Threshold-Parse: unsupported option : %s %s\n",oargs[0],oargs[1]);
         }
     }

     if( (count_flag + tracking_flag + type_flag + seconds_flag + genid_flag + sigid_flag) != 6 )
     {
	/* Fatal - incorrect argument count */
	FatalError("Threshold-Parse: incorrect argument count\n");
     }

     if( sfthreshold_create( &thdx  ) )
     {
	FatalError("Threshold-Parse: could not create a threshold object -- only one per sid=%d!\n",thdx.sig_id);
     }
}

/*

    Parse basic CIDR block  - a.b.c.d/bits

*/
static void parseCIDR( THDX_STRUCT * thdx, char * s )
{
   char        **args;
   int          nargs;

   args = mSplit( s , "/", 2, &nargs, 0 );  /* get rule option pairs */

   if( !nargs || nargs > 2  )
   {
       FatalError("Suppress-Parse: argument pairing error\n");
   }

   /*
   *   Keep IP in network order
   */
   thdx->ip_address = inet_addr( args[0] );   

   if( nargs == 2 )
   {
       int      i;
       int      nbits;
       unsigned mask;

       nbits = atoi( args[1] );
       mask  = 1 << 31;

       for( i=0; i<nbits; i++ )
       {
          thdx->ip_mask |= mask;
          mask >>= 1;
       }

       /* 
          Put mask in network order 
       */
       thdx->ip_mask = htonl(thdx->ip_mask);       
   }
   else
   {
       thdx->ip_mask = 0xffffffff; /* requires exact ip match */
   }

   /* just in case the network is not right */
   thdx->ip_address &= thdx->ip_mask;
}

/*

   suppress gen_id #, sig_id #, track by_src|by_dst, ip cidr'

*/
void ParseSFSuppress( FILE * fp, char * rule )
{

     char        **args, **oargs;
     int         nargs, noargs;
     THDX_STRUCT thdx;
     int         genid_flag=0;
     int         sigid_flag=0;
     int         i;

     memset( &thdx, 0, sizeof(THDX_STRUCT) );

     while( (*rule <= ' ') && (*rule > 0) ) rule++; /* skip whitespace */
     while( (*rule  > ' ') ) rule++;  /* skip 'suppress' */

     args = mSplit(rule,",",15,&nargs,0);  /* get rule option pairs */

     thdx.type      =  THD_TYPE_SUPPRESS;
     thdx.priority  =  THD_PRIORITY_SUPPRESS;
     thdx.ip_address=  0;  //default is all ip's- ignore this event altogether
     thdx.ip_mask   =  0;
     thdx.tracking  =  THD_TRK_DST;

     for( i=0; i<nargs; i++ )
     {
         oargs = mSplit(args[i]," ",2,&noargs,0);  /* get rule option pairs */
         if( noargs != 2 )
         {
             FatalError("Suppress-Parse: argument pairing error\n");
         }

         if( strcmp(oargs[0],"track")==0 )
         {
            if( strcmp(oargs[1],"by_src") == 0 )
            {
                thdx.tracking = THD_TRK_SRC;
            }
            else if( strcmp(oargs[1],"by_dst") == 0 )
            {
                thdx.tracking = THD_TRK_DST;
            }
            else
            {
                /* Fatal incorrect threshold type */
                 FatalError("Suppress-Parse: incorrect tracking type\n");
            }
         }

         else if( strcmp(oargs[0],"gen_id")==0 )
         {
            thdx.gen_id = atoi(oargs[1]);
            sigid_flag++;
         }

         else if( strcmp(oargs[0],"sig_id")==0 )
         {
            thdx.sig_id = atoi(oargs[1]);
            genid_flag++;
         }

         else if( strcmp(oargs[0],"ip")==0 )
         {
            parseCIDR( &thdx, oargs[1] );
         }
     }

     if( ( genid_flag + sigid_flag) != 2 )
     {
         /* Fatal - incorrect argument count */
         FatalError("Suppress-Parse: incorrect argument count\n");
     }

     if( sfthreshold_create( &thdx  ) )
     {
         FatalError("Suppress-Parse: could not create a threshold object\n");
     }
}

/*

    Init Thresholding - call when starting to parsing rules - so we can add them them

    if the init function is not called, than all thresholding is turned off because
    the thd_struct pointer is null.

*/
int sfthreshold_init()
{
   if( !s_enabled )
       return 0;

   /* Check if already init'd */
   if( s_thd )
       return 0;

   s_thd = sfthd_new( s_memcap );
   if( !s_thd )
   {
       return -1;
   }

   return 0;
}

/*
*  DEBUGGING ONLY
*/
void print_netip(unsigned ip)
{
    /* Net order has high byte in our low byte on intel - just testing anyways */
    printf("%d",(ip)&0xff);
    printf(".");
    printf("%d",(ip>>8)&0xff);
    printf(".");
    printf("%d",(ip>>16)&0xff);
    printf(".");
    printf("%d",(ip>>24)&0xff);
}

/*
*  DEBUGGING ONLY
*/
void print_thdx( THDX_STRUCT * thdx )
{
    if( thdx->type != THD_TYPE_SUPPRESS )
    {
       printf("THRESHOLD: gen_id=%d, sig_id=%d, type=%d, tracking=%d, count=%d, seconds=%d \n",
                       thdx->gen_id,
                       thdx->sig_id,
                       thdx->type,
                       thdx->tracking,
                       thdx->count,
                       thdx->seconds );
    }
    else
    {
       printf("SUPPRESS: gen_id=%d, sig_id=%d, tracking=%d, ",
                       thdx->gen_id,
                       thdx->sig_id,
                       thdx->tracking );

       printf(" ip=");
       print_netip(thdx->ip_address); 
       printf(", mask=" );
       print_netip(thdx->ip_mask); 
       printf("\n");
    }
}

/*

    Create and Add a Thresholding Event Object

*/
int sfthreshold_create( THDX_STRUCT * thdx  )
{
   if( !s_enabled )
       return 0;

   if( !s_thd )  /* Auto init - memcap must be set 1st, which is not really a problem */
   {
       sfthreshold_init();

       if( !s_thd )
           return -1;
   }

   print_thdx( thdx );

   /* Add the object to the table - */
   return sfthd_create_threshold( s_thd,
                       thdx->gen_id,
                       thdx->sig_id,
                       thdx->tracking,
                       thdx->type,
                       thdx->priority,
                       thdx->count,
                       thdx->seconds,
   		       thdx->ip_address, 
		       thdx->ip_mask ); 
}

/*

    Test an event against the threshold object table
    to determine if it should be logged.

    It will always return the same answer until sfthreshold_reset is
    called

	gen_id:
	sig_id: 
	sip:    host ordered sip
	dip:	host ordered dip
	curtime: 

    2003-05-29 cmg:

     This code is in use in fpLogEvent, CallAlertFuncs, CallLogFuncs
     and the reset function is called in ProcessPacket


    returns 1 - log
            0 - don't log

*/
int sfthreshold_test( int gen_id, int sig_id, unsigned sip, unsigned dip, long curtime )
{
   //printf("sfthreshold_test...\n");
   //fflush(stdout);
   
   if( !s_enabled )
   {
       //printf("sfthreshold_test - disabled...\n");
       return 1;
   }
  
   if( !s_thd )
   {
       //printf("sfthreshold_test - no threshold table...\n");
       return 1;
   }

   if( !s_checked )
   {
      s_checked = 1;
      s_answer  = !sfthd_test_threshold( s_thd, gen_id, sig_id, sip, dip, curtime );
      //
      // alternate returns - testing the calling software....
#ifdef XXXX
      static int x=0;
      x++;
      if( x%2 )
      {
	      s_answer=1;
      }
      else
      {
	     s_answer=0 ;
      }
#endif      
   }
       
   //if( s_answer ) printf("log it...\n");
   //else           printf("don't log it...\n");
   
   //printf("exiting sfthreshold_test...\n");
   //fflush(stdout);
   
   return s_answer;
}

/** 
 * Reset the thresholding system so that subsequent calls to
 * sfthreshold_test will indeed try to alter the thresholding system
 *
 */
void sfthreshold_reset(void)
{
    s_checked = 0;
}
